-- top module of
-- GODIL


library ieee;
use ieee.std_logic_1164.all;

library zpu;
use zpu.zpupkg.all;                     -- zpu_dbgo_t

library unisim;
use unisim.vcomponents.dcm_sp;


entity top is
    port (
        m49   : in    std_logic;        -- oscillator input
        -- GPIO
        sw1   : in    std_logic;        -- switch 1, high active
        sw2   : in    std_logic;        -- switch 2, low active
        -- TUSB3410
        sin   : inout std_logic;        -- M0 configuration pin, TUSB3410 serial data input, LED3
        sout  : in    std_logic;        -- TUSB3410 serial data out
        rts   : in    std_logic;        -- TUSB3410 ready to send (LED5)
        cts   : inout std_logic;        -- TUSB3410 clear to send (and LED6)
        vs2   : inout std_logic;        -- TUSB3410 I2C connection, LED8
        tvs1  : inout std_logic;        -- TUSB3410 I2C connector (and E2)
        -- SPI flash
        cso   : inout std_logic;        -- SPI memory chip select
        tmosi : inout std_logic;        -- SPI memory mosi (and E4)
        tdin  : inout std_logic;        -- SPI memory data out (and E5)
        tcclk : inout std_logic;        -- SPI memory clock (and E6)
        -- remaining IO pins
        c13   : in    std_logic;        -- external input (pin 49)
        d13   : in    std_logic;        -- external input (pin 50)
        tvs0  : inout std_logic;        -- E3
        tm1   : inout std_logic;        -- M1 configuration pin (and E7)
        thsw  : inout std_logic;        -- HSWAP configuration pin (and E8)
        -- I/O's for DIL / main connector
        pin   : inout std_logic_vector(48 downto 1)
        );
end entity top;


architecture rtl of top is

    ---------------------------
    -- type declarations
    type zpu_type is (zpu_small, zpu_medium);

    ---------------------------
    -- constant declarations
    constant zpu_flavour : zpu_type := zpu_small;  -- choose your flavour HERE
    --
    constant word_size_c : natural  := 32;  -- 32 bits data path
    constant addr_w_c    : natural  := 18;  -- 18 bits address space=256 kB, 128 kB I/O


    ---------------------------
    -- component declarations
    component zpu_small1 is
        generic (
            word_size  : natural   := 32;      -- 32 bits data path
            d_care_val : std_logic := '0';     -- Fill value
            clk_freq   : positive  := 50;      -- 50 MHz clock
            brate      : positive  := 115200;  -- RS232 baudrate
            addr_w     : natural   := 16;      -- 16 bits address space=64 kB, 32 kB I/O
            bram_w     : natural   := 15       -- 15 bits RAM space=32 kB
            );
        port (
            clk_i      : in  std_logic;        -- CPU clock
            rst_i      : in  std_logic;        -- Reset
            break_o    : out std_logic;        -- Break executed
            dbg_o      : out zpu_dbgo_t;       -- Debug info
            rs232_tx_o : out std_logic;        -- UART Tx
            rs232_rx_i : in  std_logic         -- UART Rx
            );
    end component zpu_small1;

    component zpu_med1 is
        generic(
            word_size  : natural   := 32;      -- 32 bits data path
            d_care_val : std_logic := '0';     -- Fill value
            clk_freq   : positive  := 50;      -- 50 MHz clock
            brate      : positive  := 115200;  -- RS232 baudrate
            addr_w     : natural   := 18;      -- 18 bits address space=256 kB, 128 kB I/O
            bram_w     : natural   := 15       -- 15 bits RAM space=32 kB
            );
        port(
            clk_i      : in  std_logic;        -- CPU clock
            rst_i      : in  std_logic;        -- Reset
            break_o    : out std_logic;        -- Break executed
            dbg_o      : out zpu_dbgo_t;       -- Debug info
            rs232_tx_o : out std_logic;        -- UART Tx
            rs232_rx_i : in  std_logic         -- UART Rx
            );
    end component zpu_med1;


    ---------------------------
    -- signal declarations
    signal dcm_sp_i0_clk0  : std_ulogic;
    signal dcm_sp_i0_clkfx : std_ulogic;
    signal clk_fb          : std_ulogic;
    signal clk             : std_ulogic;
    --
    signal reset_shift_reg : std_ulogic_vector(3 downto 0);
    signal reset_sync      : std_ulogic;
    --
    signal zpu_i0_dbg      : zpu_dbgo_t;  -- Debug info
    signal zpu_i0_break    : std_logic;


begin

    -- drive all unused outputs
    --
    -- TUSB3410
    sin   <= 'Z';  -- M0 configuration pin, TUSB3410 serial data input, LED3
    cts   <= 'Z';  -- TUSB3410 clear to send (and LED6)
    vs2   <= 'Z';  -- TUSB3410 I2C connection, LED8
    tvs1  <= 'Z';  -- TUSB3410 I2C connector (and E2)
    -- SPI flash
    cso   <= 'Z';  -- SPI memory chip select
    tmosi <= 'Z';  -- SPI memory mosi (and E4)
    tdin  <= 'Z';  -- SPI memory data out (and E5)
    tcclk <= 'Z';  -- SPI memory clock (and E6)
    -- remaining IO pins
    tvs0  <= 'Z';  -- E3
    tm1   <= 'Z';  -- M1 configuration pin (and E7)
    thsw  <= 'Z';  -- HSWAP configuration pin (and E8)
    -- I/O's for DIL / main connector
    pin   <= (others => 'Z');


    -- digital clock manager (DCM)
    -- to generate higher/other system clock frequencys
    dcm_sp_i0 : dcm_sp
        generic map (
            startup_wait   => true,     -- wait with DONE till locked
            clkfx_multiply => 2,        -- 2 for small, 3 for medium
            clkfx_divide   => 1,        -- 1 for small, 2 for medium
            clk_feedback   => "1X"
            )
        port map (
            clkin => m49,
            clk0  => dcm_sp_i0_clk0,
            clkfx => dcm_sp_i0_clkfx,
            clkfb => clk_fb
            );

    clk_fb <= dcm_sp_i0_clk0;
    clk    <= dcm_sp_i0_clkfx;


    -- reset synchronizer
    -- generate synchronous reset
    reset_synchronizer : process(clk, sw1)
    begin
        if sw1 = '1' then
            reset_shift_reg <= (others => '1');
        elsif rising_edge(clk) then
            reset_shift_reg <= reset_shift_reg(reset_shift_reg'high-1 downto 0) & '0';
        end if;
    end process;
    reset_sync <= reset_shift_reg(reset_shift_reg'high);


    -- select instance of zpu
    zpu_i0_small : if zpu_flavour = zpu_small generate
        zpu_i0 : zpu_small1
            generic map (
                addr_w    => addr_w_c,
                word_size => word_size_c,
                clk_freq  => 98
                )
            port map (
                clk_i      => clk,           -- : in  std_logic;   -- CPU clock
                rst_i      => reset_sync,    -- : in  std_logic;   -- Reset
                break_o    => zpu_i0_break,  -- : out std_logic;   -- Break executed
                dbg_o      => zpu_i0_dbg,    -- : out zpu_dbgo_t;  -- Debug info
                rs232_tx_o => sin,           -- : out std_logic;   -- UART Tx
                rs232_rx_i => sout           -- : in  std_logic    -- UART Rx
                );
    end generate zpu_i0_small;

    zpu_i0_medium : if zpu_flavour = zpu_medium generate
        zpu_i0 : zpu_med1
            generic map (
                addr_w    => addr_w_c,
                word_size => word_size_c,
                --clk_freq  => 75
                clk_freq  => 49
                )
            port map (
                clk_i      => clk,           -- : in  std_logic;   -- CPU clock
                rst_i      => reset_sync,    -- : in  std_logic;   -- Reset
                break_o    => zpu_i0_break,  -- : out std_logic;   -- Break executed
                dbg_o      => zpu_i0_dbg,    -- : out zpu_dbgo_t;  -- Debug info
                rs232_tx_o => sin,           -- : out std_logic;   -- UART Tx
                rs232_rx_i => sout           -- : in  std_logic    -- UART Rx
                );
    end generate zpu_i0_medium;

    -- pragma translate_off 
    trace_mod : trace
        generic map (
            addr_w    => addr_w_c,
            word_size => word_size_c,
            log_file  => "zpu_trace.log"
            )
        port map (
            clk_i  => clk,
            dbg_i  => zpu_i0_dbg,
            stop_i => zpu_i0_break,
            busy_i => '0'
            );
    -- pragma translate_on

    -- switch on all LEDs in case of break
    process
    begin
        wait until rising_edge(clk);
        if zpu_i0_break = '1' then
            vs2 <= '1';
        end if;
        if reset_sync = '1' then
            vs2 <= '0';
        end if;
    end process;

end architecture rtl;
